/* TODO
 * set correct window before event-handling
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <X11/Xlib.h>
#include <GL/glx.h>
#include "glass.h"

#define MAX_ATTRIB	32
#define MAX_WIN		32

#define FOODEBUG \
	XSync(dpy, False); printf("skata\n");

struct window {
	int used;

	int width, height;
	int fullscr;
	int rsize, gsize, bsize, asize, zsize, ssize;

	Window win;
	GLXContext ctx;

	int mapped;
	unsigned int inp_mask;

	void (*callbacks[GLA_MAX_CB])();
};

static int attr_gla2glx(int attr);

static Display *dpy;
static struct window wins[MAX_WIN];
static int cur_win = -1;
static Atom xa_wm_prot, xa_wm_del;

int gla_create_win(int xsz, int ysz, struct gla_visual *gv)
{
	struct window win;
	Window root;
	int i, scr, widx, prev_cur_win;
	XVisualInfo *vis;
	XSetWindowAttributes xattr;
	XClassHint class_hint;
	static struct gla_visual def_gv = {0, {GLX_RGBA, GLX_DOUBLEBUFFER,
		GLX_RED_SIZE, 8, GLX_GREEN_SIZE, 8, GLX_BLUE_SIZE, 8,
		GLX_DEPTH_SIZE, 24, None}};

	widx = -1;
	for(i=0; i<MAX_WIN; i++) {
		if(!wins[i].used) {
			widx = i;
			break;
		}
	}
	if(widx == -1) {
		fprintf(stderr, "can't create more windows\n");
		return -1;
	}
	

	if(!dpy) {
		if(!(dpy = XOpenDisplay(0))) {
			return -1;
		}
		xa_wm_prot = XInternAtom(dpy, "WM_PROTOCOLS", False);
		xa_wm_del = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
	}
	scr = DefaultScreen(dpy);
	root = RootWindow(dpy, scr);

	if(!gv) {
		gv = &def_gv;
	}
	

	if(!(vis = glXChooseVisual(dpy, scr, gv->attr))) {
		fprintf(stderr, "couldn't find suitable visual.\n");
		return -1;
	}

	glXGetConfig(dpy, vis, GLX_RED_SIZE, &win.rsize);
	glXGetConfig(dpy, vis, GLX_GREEN_SIZE, &win.gsize);
	glXGetConfig(dpy, vis, GLX_BLUE_SIZE, &win.bsize);
	glXGetConfig(dpy, vis, GLX_ALPHA_SIZE, &win.asize);
	glXGetConfig(dpy, vis, GLX_DEPTH_SIZE, &win.zsize);
	glXGetConfig(dpy, vis, GLX_STENCIL_SIZE, &win.ssize);

	if(!(win.ctx = glXCreateContext(dpy, vis, 0, True))) {
		fprintf(stderr, "failed to create GLX context\n");
		XFree(vis);
		return -1;
	}

	xattr.background_pixel = xattr.border_pixel = BlackPixel(dpy, scr);
	xattr.colormap = XCreateColormap(dpy, root, vis->visual, AllocNone);

	if(!(win.win = XCreateWindow(dpy, root, 0, 0, xsz, ysz, 0, vis->depth, InputOutput,
					vis->visual, CWColormap | CWBackPixel | CWBorderPixel, &xattr))) {
		fprintf(stderr, "failed to create X window\n");
		glXDestroyContext(dpy, win.ctx);
		XFree(vis);
		return -1;
	}
	XFree(vis);

	win.width = xsz;
	win.height = ysz;
	win.fullscr = 0;

	wins[widx] = win;
	wins[widx].used = 1;
	prev_cur_win = cur_win;
	cur_win = widx;

	XSetWMProtocols(dpy, win.win, &xa_wm_del, 1);
	gla_win_title("glass window");

	class_hint.res_name = "glass";
	class_hint.res_class = "glass";
	XSetClassHint(dpy, win.win, &class_hint);

	if(prev_cur_win != -1) {
		cur_win = prev_cur_win;
	}

	glXMakeCurrent(dpy, win.win, win.ctx);
	XMapWindow(dpy, win.win);
	return widx;
}

void gla_destroy_win(void)
{
	glXDestroyContext(dpy, wins[cur_win].ctx);
	XDestroyWindow(dpy, wins[cur_win].win);
	wins[cur_win].used = 0;
}

void gla_show_win(void)
{
	XMapWindow(dpy, wins[cur_win].win);
}

void gla_hide_win(void)
{
	XUnmapWindow(dpy, wins[cur_win].win);
}

void gla_win_title(const char *title)
{
	XTextProperty wm_name;

	XStringListToTextProperty((char**)&title, 1, &wm_name);
	XSetWMName(dpy, wins[cur_win].win, &wm_name);		/* set WM_NAME property */
	XSetWMIconName(dpy, wins[cur_win].win, &wm_name);	/* set WM_ICON_NAME property */
	XFree(wm_name.value);
}

void gla_fullscreen(void)
{
	/* TODO */
}

void gla_windowed(void)
{
	/* TODO */
}

#define HAVE_VALUE(arg)	((arg) >= GLA_RED)
void gla_visual(struct gla_visual *gv, ...)
{
	va_list ap;
	int arg;
	int idx = 0;

	va_start(ap, gv);

	do {
		arg = va_arg(ap, int);

		if(arg == GLA_FULLSCREEN) {
			gv->fullscr = 1;
			continue;
		}
		
		gv->attr[idx++] = attr_gla2glx(arg);

		if(HAVE_VALUE(arg)) {
			gv->attr[idx++] = va_arg(ap, int);
		}
	} while(arg);
	
	va_end(ap);
}

void gla_callback(int cbidx, void (*func)())
{
	unsigned int mask;
	wins[cur_win].callbacks[cbidx] = func;

	switch(cbidx) {
	case GLA_EXPOSE:
		mask = ExposureMask;
		break;

	case GLA_RESHAPE:
		mask = StructureNotifyMask;
		break;

	case GLA_KEYBOARD:
		mask = KeyPressMask | KeyReleaseMask;
		break;

	case GLA_BUTTON:
		mask = ButtonPressMask | ButtonReleaseMask;
		break;

	case GLA_MOTION:
		mask = ButtonMotionMask;
		break;

	case GLA_PASSIVE_MOTION:
		mask = PointerMotionMask;
		break;

	default:
		return;
	}

	if(func) {
		wins[cur_win].inp_mask |= mask;
	} else {
		wins[cur_win].inp_mask &= ~mask;
	}
	XSelectInput(dpy, wins[cur_win].win, wins[cur_win].inp_mask);
}

void gla_redisplay(void)
{
	XEvent xev;
	xev.type = Expose;
	xev.xexpose.window = wins[cur_win].win;
	xev.xexpose.x = xev.xexpose.y = 0;
	xev.xexpose.width = wins[cur_win].width;
	xev.xexpose.height = wins[cur_win].height;
	xev.xexpose.count = 0;

	XSendEvent(dpy, wins[cur_win].win, True, 0, &xev);
}

void gla_swap_buffers(void)
{
	glXSwapBuffers(dpy, wins[cur_win].win);
}

int gla_event_proc(void)
{
	XEvent xev;
	struct window *win = wins + cur_win;
	static int bn_pressed;

	if(win->callbacks[GLA_IDLE] && !XPending(dpy)) {
		win->callbacks[GLA_IDLE]();
		return 1;
	}

	XNextEvent(dpy, &xev);

	switch(xev.type) {
	case MapNotify:
		win->mapped = 1;
		break;

	case UnmapNotify:
		win->mapped = 0;
		break;

	case ConfigureNotify:
		if(xev.xconfigure.width != win->width || xev.xconfigure.height != win->height) {
			win->width = xev.xconfigure.width;
			win->height = xev.xconfigure.height;

			if(win->callbacks[GLA_RESHAPE]) {
				win->callbacks[GLA_RESHAPE](win->width, win->height);
			}
		}
		break;

	case Expose:
		if(win->mapped && xev.xexpose.count == 0 && win->callbacks[GLA_EXPOSE]) {
			win->callbacks[GLA_EXPOSE]();
		}
		break;

	case ClientMessage:
		if(xev.xclient.message_type == xa_wm_prot) {
			if(xev.xclient.data.l[0] == xa_wm_del) {
				return 0;
			}
		}
		break;

	case KeyPress:
	case KeyRelease:
		if(win->callbacks[GLA_KEYBOARD]) {
			KeySym sym = XLookupKeysym((XKeyEvent*)&xev.xkey, 0);
			win->callbacks[GLA_KEYBOARD]((int)sym & 0xff, xev.type == KeyPress ? 1 : 0);
		}
		break;

	case ButtonPress:
		bn_pressed++;
		if(0) {
	case ButtonRelease:
			bn_pressed--;
		}
		if(win->callbacks[GLA_BUTTON]) {
			win->callbacks[GLA_BUTTON](xev.xbutton.button - 1, xev.type == ButtonPress ? 1 : 0,
					xev.xbutton.x, xev.xbutton.y);
		}
		break;

	case MotionNotify:
		if(bn_pressed > 0) {
			if(win->callbacks[GLA_MOTION]) {
				win->callbacks[GLA_MOTION](xev.xmotion.x, xev.xmotion.y);
			}
		} else {
			if(win->callbacks[GLA_PASSIVE_MOTION]) {
				win->callbacks[GLA_PASSIVE_MOTION](xev.xmotion.x, xev.xmotion.y);
			}
		}
		break;

	default:
		break;
	}
	return 1;
}

void gla_main_loop(void)
{
	/* TODO */
}

void *gla_get(int what)
{
	switch(what) {
	case GLA_WIDTH:
		return &wins[cur_win].width;
	case GLA_HEIGHT:
		return &wins[cur_win].height;
	case GLA_DISPLAY:
		return dpy;
	case GLA_WINDOW:
		return &wins[cur_win].win;
	default:
		break;
	}
	return 0;
}

static int attr_gla2glx(int attr)
{
	switch(attr) {
	case GLA_RGBA:
		return GLX_RGBA;
	case GLA_DOUBLE:
		return GLX_DOUBLEBUFFER;
	case GLA_STEREO:
		return GLX_STEREO;
	case GLA_RED:
		return GLX_RED_SIZE;
	case GLA_GREEN:
		return GLX_GREEN_SIZE;
	case GLA_BLUE:
		return GLX_BLUE_SIZE;
	case GLA_ALPHA:
		return GLX_ALPHA_SIZE;
	case GLA_DEPTH:
		return GLX_DEPTH_SIZE;
	case GLA_STENCIL:
		return GLX_STENCIL_SIZE;
	default:
		break;
	}
	return None;
}
