#include <stdio.h>
#include <stdlib.h>
#include <GL/gl.h>
#include <GL/glut.h>
#include "glass.h"
#include "cam.h"


void disp(void);
void render_scene(float tsec);
void reshape(int x, int y);
void keyb(int key, int state);
void mouse(int bn, int state, int x, int y);
void motion(int x, int y);

int use_stereo;

int main(int argc, char **argv)
{
	struct gla_visual gvis;

	glutInit(&argc, argv);

	gla_visual(&gvis, GLA_RGBA, GLA_DOUBLE, GLA_STEREO,
			GLA_RED, 5, GLA_GREEN, 5, GLA_BLUE, 5,
			GLA_NONE);

	if(gla_create_win(640, 480, &gvis) == -1) {
		fprintf(stderr, "gla_create_win failed\n");
		return 1;
	}

	gla_callback(GLA_EXPOSE, disp);
	gla_callback(GLA_RESHAPE, reshape);
	gla_callback(GLA_KEYBOARD, keyb);
	gla_callback(GLA_BUTTON, mouse);
	gla_callback(GLA_MOTION, motion);
	gla_callback(GLA_IDLE, disp);

	/*glEnable(GL_DEPTH_TEST);*/
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	{
		float lpos[] = {-1, 1, 3, 0};
		glLightfv(GL_LIGHT0, GL_POSITION, lpos);
	}

	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	cam_focus_dist(10.0);

	while(gla_event_proc());

	printf("exiting\n");
	return 0;
}

void disp(void)
{
	unsigned long msec = 0;/*glutGet(GLUT_ELAPSED_TIME);*/
	float sec = msec / 1000.0;

	float col[] = {1, 0.3, 0.2, 1};
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, col);

	if(!use_stereo) {
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		cam_proj_matrix();

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		cam_view_matrix();
	
		glDrawBuffer(GL_BACK_LEFT);
		render_scene(sec);
		glDrawBuffer(GL_BACK_RIGHT);
		render_scene(sec);
	} else {

		/* render left eye */
		glDrawBuffer(GL_BACK_LEFT);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		cam_stereo_proj_matrix(CAM_LEFT);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		cam_stereo_view_matrix(CAM_LEFT);

		render_scene(sec);

		/* render right eye */
		glDrawBuffer(GL_BACK_RIGHT);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		cam_stereo_proj_matrix(CAM_RIGHT);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		cam_stereo_view_matrix(CAM_RIGHT);

		render_scene(sec);
	}

	gla_swap_buffers();
}

void render_scene(float tsec)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glTranslatef(0, 0.5, 0);

	glFrontFace(GL_CW);
	glutSolidTeapot(1.0);
	glFrontFace(GL_CCW);

	glPopMatrix();
}

void reshape(int x, int y)
{
	glViewport(0, 0, x, y);
}

void keyb(int key, int state)
{
	if(state == 0) return;

	switch(key) {
	case 27:
		exit(0);

	case 's':
	case 'S':
		use_stereo = !use_stereo;
		/*gla_redisplay();*/
		break;

	default:
		break;
	}
}

static int prev_x, prev_y;
static int bnstate[16];

void mouse(int bn, int state, int x, int y)
{
	if(state) {
		prev_x = x;
		prev_y = y;
	} else {
		prev_x = -1;
	}
	bnstate[bn] = state;
}

void motion(int x, int y)
{
	if(bnstate[0]) {
		cam_rotate(x - prev_x, y - prev_y);
		/*gla_redisplay();*/
	}
	if(bnstate[1]) {
		cam_pan(x - prev_x, y - prev_y);
		/*gla_redisplay();*/
	}
	if(bnstate[2]) {
		cam_zoom(y - prev_y);
		/*gla_redisplay();*/
	}
	prev_x = x;
	prev_y = y;
}
